from __future__ import division
from collections import namedtuple, OrderedDict
import six
import numpy as np
import xarray as xr
import logging
from itertools import chain
import iucm.energy_consumption as en
from iucm.utils import docstrings
import iucm.utils as utils
import psyplot.data as psyd


logger = logging.getLogger(__name__)


def _nothing():
    return None, None


#: Meta information for the variables in the 1D-output of the
#: :class:`PopulationModel`
fields = OrderedDict([
    ('cons', {
        'units': 'MJ / inh',
        'long_name': 'Energy consumption',
        }),
    ('dist', {
        'units': 'km',
        'long_name': 'Average distance between two individuals',
        }),
    ('entrop', {
        'units': '-',
        'long_name': 'Entropy',
        }),
    ('rss', {
        'units': '-',
        'long_name': 'Rank-Size Slope',
        }),
    ('cons_det', {
        'units': 'MJ / inh',
        'long_name': 'Deterministic energy consumption',
        'comments': ('Energy consumption calculated based on based on '
                     'Le Nechet, 2012'),
        }),
    ('cons_std', {
        'units': 'MJ / inh',
        'long_name': 'std. deviation of Energy consumption',
        }),
    ('left_over', {
        'units': 'inhabitants',
        'long_name': 'Left over population that could not be subtracted',
        }),
    ('nscenarios', {
        'units': '-',
        'long_name': (
            'The number of scenarios that have been changed during this step'),
        }),
    ])

Output = utils.append_doc(
    namedtuple('Output', list(fields)),
    """The state of the model

    The :func:`collections.namedtuple` defining the state of the model during
    a single step. Each field of this class corresponds to one output variable
    in the output netCDF of the :class:`PopulationModel`. Meta information are
    taken from the :attr:`fields` dictionary

    Parameters
    ----------
    cons: float
        Energy consumption
    dist: float
        Average distance between two individuals
    entrop: float
        Entropy
    rss: float
        Rank-Size Slope
    cons_det: float
        The deterministic energy consumption based on
        :attr:`iucm.energy_consumption.weights_LeNechet`
    cons_std: float
        The standard deviation of the energy consumption
    left_over: float
        The left over inhabitants that could not be subtracted in the last step
    nscenarios: float
        The number of scenarios that have been changed during this step

    See Also
    --------
    Output2D
    PopulationModel.state
    PopulationModel.allocate_output
    """)


#: Meta information for the variables in the 2D-output of the
#: :class:`PopulationModel`
fields2D = fields.copy()
fields2D['scenarios'] = {'long_name': 'Scenario identifier'}


Output2D = utils.append_doc(
    namedtuple('Output2D', list(fields2D)),
    """The 2D state variables of the model

    2D output variables. Each variable corresponds to the same variable as in
    0-d :class:`Output` objects, but instead of saving only the best state of
    the model, this output saves all scenarios

    Parameters
    ----------
    cons: np.ndarray of dtype float
        Energy consumption
    dist: np.ndarray of dtype float
        Average distance between two individuals
    entrop: np.ndarray of dtype float
        Entropy
    rss: np.ndarray of dtype float
        Rank-Size Slope
    cons_det: np.ndarray of dtype float
        The deterministic energy consumption based on
        :attr:`iucm.energy_consumption.weights_LeNechet`
    cons_std: np.ndarray of dtype float
        The standard deviations within each probabilistic scenario
    left_over: float
        The left over inhabitants that could not be subtracted in the last step
    nscenarios: float
        The weight of each scenario that has been used during this step
    scenarios: np.ndarray of dtype int
        The number of the scenario

    See Also
    --------
    Output
    PopulationModel.state2d""")


class PopulationModel(object):
    """
    Class that runs and manages the population model

    This class represents one instance of the model for one experiment and is
    responsible for all the computation, parallelization and input-output
    coordination.
    The major important features of this class are

    :meth:`from_da` method
        A classmethod to construct the model from a :class:`xarray.DataArray`
    :meth:`step` method
        The method that is brings the model to the next step
    :attr:`init_step_methods` and :attr:`update_methods`
        The available update methods how to bring the model to the next change
    :attr:`selection_methods`
        The available selection methods that define the available scenarios
    :attr:`update_methods`
        The available update methods that define how the population change
        for the given selection is pursued
    :attr:`state` attribute
        The current state of the model. It's an instance of the :class:`Output`
        class containing all 1D-variables of the current :attr:`data` attribute

    Most of the other routines are related to input/output and parallelization
    """

    @property
    def dist(self):
        """Denumerator of average distance between two individuals"""
        return self.state.dist

    @property
    def consumption(self):
        """Energy consumption"""
        return self.state.cons

    @property
    def population_change(self):
        """The population change during this time step"""
        return self._change[self.total_step]

    @property
    def movement(self):
        """The population movement during this time step"""
        return self._movement[self.total_step]

    @property
    def total_change(self):
        """The total population change during this time step"""
        return self.movement + self.population_change + self.left_over

    @property
    def state(self):
        """The state as a namedtuple. You may set it with an iterable defined
        by the :attr:`Output` class"""
        return self._state

    @state.setter
    def state(self, value):
        self._state = Output(*value)

    @property
    def state2d(self):
        """The different values from the :attr:`state` of the model for each
        of the scenarios"""
        return self._state2d

    @state2d.setter
    def state2d(self, value):
        self._state2d = Output2D(*value)

    @property
    def state_dict(self):
        """The state as a dictionary. Mapping from state variable to the
        corresponding value. You may also set it with a dictionary"""
        return self._state._asdict()

    @state_dict.setter
    def state_dict(self, kwargs):
        self._state = Output(**kwargs)

    @property
    def state2d_dict(self):
        """The state as a dictionary. Mapping from state variable to the
        corresponding value. You may also set it with a dictionary"""
        return self._state2d._asdict()

    @state2d_dict.setter
    def state2d_dict(self, kwargs):
        self._state2d = Output(**kwargs)

    @property
    def left_over(self):
        """Left over population that could not have been subtracted for within
        the :meth:`PopulationModel.value_update` method and should be
        considered in the next step"""
        return self.state.left_over

    #: Left over population that could not have been subtracted for within the
    #: :meth:`PopulationModel.value_update` method
    _left_over = 0

    #: :class:`int`. The number of cells that are modified for one scenario
    ncells = 4

    #: The current step since the last output. -1 means, output has just been
    #: written or the model has been initialized
    current_step = -1

    #: The current step since the initialization of the model. -1 means, the
    #: model has been initialized
    total_step_run = -1

    #: The absolute current step in case the model has been restarted. If -1,
    #: the model has not yet been started
    total_step = -1

    #: np.ndarray. The current simulation data of the model
    data = None

    #: np.ndarray. The x-coordinates of the points in :attr:`data`
    x = None

    #: np.ndarray. The y-coordinates of the points in :attr:`data`
    y = None

    #: np.ndarray. The positions of the points in :attr:`data` that can be
    #: modified
    data2modify = None

    #: Attribute that is either False or the name of the output file where the
    #: model data just has been written to
    output_written = False

    #: The selection method from :attr:`selection_methods` we use to define the
    #: scenarios
    select = None

    #: The step initialization method from :attr:`init_step_methods` we use to
    #: define the scenarios
    init_step = None

    #: The update method from :attr:`update_methods` we use to compute that
    #: changes for each scenario
    update = None

    @docstrings.get_sectionsf('PopulationModel')
    def __init__(self, data, x, y, selection_method='consecutive',
                 update_method='categorical', ncells=4, categories=None,
                 state=None, forcing=None, probabilistic=0, max_pop=None,
                 use_pctls=False, last_step=0, data2modify=None):
        """
        Parameters
        ----------
        data: np.ndarray
            The 1D-data array (without NaN!) of the model
        x: np.ndarray
            The x-coordinates in *km* of each point in `data` (same shape as
            `data`)
        y: np.ndarray
            The y-coordinates in *km* of each point in `data` (same shape as
            `data`)
        selection_method: { 'consecutive' | 'random' }
            The avaiable selection scenarios (see :attr:`selection_methods`)
        update_method: { 'categorical' | 'random' | 'forced' }
            The avaiable update methods (see :attr:`update_methods`)
        ncells: int
            The number of cells that shall be modified for one scenario. The
            higher the number, the less computationally expensive is the
            computation
        categories: list of str
            The categories to use. If `update_method` is ``'categorical'``, it
            describes the categories and if `use_pctls` is True, it the
            each category is interpreted as a quantile in `data`
        state: list of float
            The current state of `data`. Must be a list corresponding to the
            :class:`Output` class
        forcing: xarray.Dataset
            The input dataset for the model containing variables with
            population evolution information. Possible variables in the netCDF
            file are *movement* containing the number of people to move and
            *change* containing the population change (positive or negative)
        probabilistic: int or tuple
            The number of probabilistic scenarios. For each scenario the energy
            consumption is calculated and the final population is distributed
            to the cells with the ideal energy consumption. Set this to 0 to
            only use the weights by [LeNechet2012]_. If tuple, then they are
            considered as the weights
        max_pop: np.ndarray
            A 1d-array with the maximum population for each cell in `data`. If
            None, the last value in `categories` will be used
        use_pctls: bool
            If True, values given in `categories` are interpreted as quantiles
        last_step: int
            If the model is restarted, the total number of already made steps
            (see :attr:`total_step` attribute)
        data2modify: np.ndarray
            The indices of points in `data` which are allowed to be modified.
            If None, all points are allowed to be modified

        See Also
        --------
        from_da: A more convenient initialization method using a xarray.Dataset
        """
        self.data = data
        self.x = x
        self.y = y
        if data2modify is not None:
            data2modify = data2modify.astype(np.int64)
        self.data2modify = data2modify

        self.selection_method = selection_method

        # set the maximum population
        max_pop = np.asarray(max_pop)
        if max_pop.ndim == 0:
            if not max_pop and categories is None:
                raise ValueError(
                    "The categories must be specified if max_pop is None!")
            elif not max_pop:  # max_pop is None
                max_pop = categories[-1]
            else:
                categories = np.append(categories, max_pop)
            self.max_pop = np.empty_like(data)
            self.max_pop[:] = max_pop
        else:
            assert max_pop.shape == data.shape, (
                "Shape of max_pop must %r equal the shape of the data %r" % (
                    max_pop.shape, data.shape))
            self.max_pop = max_pop

        if use_pctls:
            categories = np.percentile(
                self.data[~np.isnan(self.data) & (self.data > 0)],
                categories)
        self.categories = np.asarray(categories)
        try:
            self.select = self.selection_methods[selection_method]
        except KeyError:
            raise ValueError(
                "Unknown selection method %s! Possibilities are {%s}" % (
                    selection_method, ', '.join(self.selection_methods)))
        self.ncells = ncells
        self.update_method = update_method
        try:
            self.update = self.update_methods[update_method]
        except KeyError:
            raise ValueError(
                "Unknown update method %s! Possibilities are {%s}" % (
                    update_method, ', '.join(self.update_methods)))
        self.init_step = self.init_step_methods[update_method]
        if update_method == 'forced' and forcing is None:
            raise ValueError(
                "Need forcing information for forced update method! Please "
                "use the forcing keyword!")
        # determine the weights
        self.nprobabilistic = probabilistic
        if probabilistic:
            if update_method != 'forced':
                raise ValueError(
                    "Can only do probabilistic runs for a forced model! "
                    "Set probabilistic to 0 or update_method to 'forced'!")
            self.weights = en.EnVariables(
                *(np.zeros(probabilistic) for var in en.EnVariables._fields))
        else:  # use the weights from LeNechet, 2012
            self.weights = en.EnVariables(
                *np.reshape(en.weights_LeNechet,
                            (len(en.EnVariables._fields), 1)))
        # set the state
        if state is None:
            # calculate data
            logger.debug('Calculating state')
            en_info = en.energy_consumption(self.data, self.x, self.y,
                                            weights=self.weights)
            self.state = Output(
                en_info[0].mean(), *en_info[1:],
                cons_det=en._calculate_en(en_info[1:]),
                cons_std=en_info[0].std(), left_over=0, nscenarios=0)
            logger.debug('Done')
        else:
            self.state = state
        self.state2d = [np.zeros_like(self.data) for _ in range(
                            len(self.state) + 1)]
        for arr in self.state2d:
            arr[:] = np.nan
        self.forcing = forcing
        if forcing is not None:
            self._change = forcing.variables['change'].values
            self._movement = forcing.variables['movement'].values
        self.current_step = -1
        self.total_step = last_step - 1
        self.total_step_run = -1

    @docstrings.get_sectionsf('PopulationModel.initialize_model')
    @docstrings.dedent
    def initialize_model(self, da, dsi, ofiles, osteps, mask=None):
        """
        Initialize the model on the I/O processor

        Parameters
        ----------
        data: xr.DataArray
            The dataarray during the initialization
        dsi: xr.Dataset
            The base dataset of the `da`
        ofiles: list of str
            The name of the output files
        osteps: list of int
            Steps when to make the output
        mask: np.ndarray
            A boolean array that maps from the :attr:`data` attribute into the
            2D output data array
        """
        if isinstance(ofiles, six.string_types):
            ofiles = [ofiles]
        ofiles = iter(ofiles)
        try:
            osteps = iter(osteps)
        except TypeError:
            osteps = iter([osteps])
        self._ofiles = ofiles
        self._osteps = osteps
        self._next_ostep = next(self._osteps)
        if mask is None:
            mask = np.zeros(da.shape, dtype=bool)
        self.mask = mask
        self.allocate_output(da, steps=self._next_ostep, dsi=dsi)

    docstrings.delete_params('PopulationModel.parameters',
                             'data', 'x', 'y', 'data2modify')
    docstrings.delete_params('PopulationModel.initialize_model.parameters',
                             'mask')

    @classmethod
    @docstrings.dedent
    def from_da(cls, da, dsi, ofiles=None, osteps=None,
                coord_transform=1, **kwargs):
        """
        Construct the model from a :class:`psyplot.data.InteractiveArray`

        Parameters
        ----------
        %(PopulationModel.initialize_model.parameters.no_mask)s
        coord_transform: float
            The transformation factor to transform the coordinate values into
            kilometres

        Other Parameters
        ----------------
        %(PopulationModel.parameters.no_data|x|y|data2modify)s

        Returns
        -------
        PopulationModel
            The model created ready to use"""
        def get_val(name):
            var = dsi.variables[name]
            return var.values[() if var.ndim == 0 else -1]
        current = da if da.ndim < 3 else da[-1]
        data = current.values.ravel().copy()
        x = current.coords[current.dims[-1]].values
        y = current.coords[current.dims[-2 if current.ndim > 1 else -1]].values
        if x.shape != data.shape:
            x, y = [arr.ravel() for arr in np.meshgrid(x, y)]
        # apply the mask to filter out NaN values
        mask = ~np.isnan(data)
        data = data[mask].ravel().astype(np.float64)
        x = x[mask.ravel()]
        y = y[mask.ravel()]
        if not set(fields).difference(dsi.variables):
            state = map(get_val, fields)
        else:
            state = None

        # check for values that should not be modified
        if 'mask' in dsi.variables and 'data2modify' not in kwargs:
            modi_mask = dsi.mask.values.ravel()[mask].astype(bool)
            data2modify = np.arange(data.size, dtype=np.int64)[~modi_mask]
            kwargs['data2modify'] = data2modify
        if 'max_pop' in dsi.variables and not kwargs.get('max_pop'):
            kwargs['max_pop'] = dsi.variables['max_pop'].values.ravel()[
                mask.ravel()]

        ret = cls(data, x * coord_transform, y * coord_transform, state=state,
                  **kwargs)
        # initialize the model and allocate the output
        if ofiles is not None:
            ret.initialize_model(da, ofiles=ofiles, osteps=osteps, dsi=dsi,
                                 mask=mask.reshape(current.shape))
        return ret

    def best_scenario(self, all_slices, all_indices):
        """
        Compute the best scenario

        This method computes the best scenario for the given scenarios defined
        through the given `slices` and `indices`

        Parameters
        ----------
        slices: list of ``None``, :class:`slice` or boolean arrays
            The slicing objects for each scenario that allow us to create a
            view of the :attr:`data` attribute that we modify in place. If list
            of ``None``, it is computed using `indices`
        indices: list of list of :class`int`
            The numpy array containing the integer position in :attr:`data` of
            the cells modified for each scenario

        Returns
        -------
        1-dim np.ndarray of dtype float
            The consumptions of the best scenarios for each set of weights used
        2-dim np.ndarray of dtype float with shape ``(nprob, len(self.state))``
            The state of the best scenario for each probabilistic scenario
            which can be used for the :attr:`state` attribute.
        list of :class:`slice` or boolean array
            The slicing object from `slices` that corresponds to the best
            scenario and can be used to create a view on the :attr:`data`
        list of list of float
            The numbers of the modified cells for the best scenario
        list of 2d-np.ndarrays
            The 2d variables of the :attr:`state2d` attribute
        """
        self.current_step += 1
        self.total_step += 1
        self.total_step_run += 1
        data = self.data
        x = self.x
        y = self.y
        update = self.update
        to_slice = self._bool_or_slice
        dist = self.dist
        nprobabilistic = self.nprobabilistic or 1
        consumptions = np.empty((len(all_slices), nprobabilistic))
        consumptions.fill(np.inf)
        others = np.empty((len(all_slices), len(self.state) - 1))
        ret_cells = [[] for _ in range(len(all_slices))]
        state2d = self.state2d
        weights = self.weights
        i_left_over = state2d._fields.index('left_over') - 1
        i_cons_std = state2d._fields.index('cons_std') - 1
        i_cons_det = state2d._fields.index('cons_det') - 1
        for arr in state2d:
            arr[:] = np.nan
        # loop through the scenarios
        for i, (sl, ind) in enumerate(zip(all_slices, all_indices)):
            all_slices[i] = sl = to_slice(data, sl, ind)
            #: store the old values
            old = data[sl].copy()
            #: The cells that are modified for this scenario
            cell_values = update(old.copy(), sl)
            #: compute the energy consumption for this scenario if (if we have
            #: any changes to the initial state)
            if (old != cell_values).any():
                en_info = en.energy_consumption(
                    data, x, y, indices=ind, increase=cell_values - old,
                    dist0=dist, slicer=sl, weights=weights)
                consumptions[i] = en_info[0]
                others[i][:len(en_info) - 1] = en_info[1:]
                others[i][i_cons_det] = en._calculate_en(en_info[1:])
                others[i][i_cons_std] = en_info[0].std()
                others[i][i_left_over] = self._left_over
                ret_cells[i] = cell_values
                state2d[0][sl] = consumptions[i].mean()
                for j, val in enumerate(others[i], 1):
                    state2d[j][sl] = val
                state2d.scenarios[sl] = i
        # get the index of the scenario with the least energy consumption
        argmins = consumptions.argmin(axis=0)
        others[:, state2d._fields.index('nscenarios') - 1] = np.unique(
            argmins).size
        return (
            # the consumptions (1-d array)
            consumptions[argmins, np.arange(nprobabilistic)],
            # the other state variables
            others[argmins],
            # the slicers
            [all_slices[i] for i in argmins],
            # the cell values
            [ret_cells[i] for i in argmins],
            # the 2d state
            state2d)

    def step(self):
        """
        Bring the model to the next step and eventually write the output

        This method is the core of the entire :class:`PopulationModel` API
        connecting the necessary functions to compute the next best scenario.
        The general structure is

        1. initialize the step (see :attr:`init_step_methods`)
        2. define the scenarios (see :attr:`selection_methods`)
        3. choose the best scenario (see :attr:`best_scenario` and
           :attr:`update_methods`)
        4. write the output (see :meth:`write` method)

        Depending on whether the :meth:`start_processes` method has been called
        earlier, this is either done serial or in parallel"""
        self.output_written = False

        # ---- initialize the step
        current = self.data.copy()
        init_sl, indices = self.init_step()
        if init_sl is not None:
            en_info = en.energy_consumption(
                self.data, self.x, self.y, indices=indices,
                dist0=self.dist, slicer=init_sl,
                increase=self.data[init_sl] - current[init_sl],
                weights=self.weights)
            self.state = Output(en_info[0].mean(), *en_info[1:],
                                cons_det=en._calculate_en(en_info[1:]),
                                cons_std=en_info[0].std(),
                                left_over=self._left_over, nscenarios=0)
        nprobabilistic = self.nprobabilistic

        # ---- define the scenarios
        all_slices, all_indices = self.select()

        if nprobabilistic:
            en.random_weights(self.weights)
        else:
            nprobabilistic = 1

        if self.procs is None:
            # ---- get the best scenario in serial
            consumptions, others, sl, changed_values, self.state2d = \
                self.best_scenario(all_slices, all_indices)
            nscenarios = others[
                0, self.state2d._fields.index('nscenarios') - 1]
        else:
            # ---- get the best scenario in parallel
            self.current_step += 1
            self.total_step += 1
            self.total_step_run += 1
            for conn in self.parent_conns:
                conn.send(['update', init_sl,
                           self.data[init_sl] if init_sl is not None else None,
                           self.state, self.weights])
                conn.recv()
            nprocs = self.nprocs
            splitted_slices = np.array_split(all_slices, nprocs)
            splitted_indices = np.array_split(all_indices, nprocs)
            for i, (conn, slices, indices) in enumerate(zip(
                    self.parent_conns, splitted_slices, splitted_indices)):
                if len(slices):
                    conn.send(['best', slices, indices])
                else:
                    nprocs = i
                    break
            #: The states of the model of the best scenario for each process
            all_consumptions = np.zeros((nprocs, nprobabilistic))
            all_others = np.zeros((nprocs, nprobabilistic,
                                   len(self.state) - 1))
            #: The slicers for the best scenario for each process
            proc_slices = [[0] * nprobabilistic] * nprocs
            #: The values of the changed cells of the best scenario for each
            #: process
            all_changed = [
                [[] for _ in range(nprobabilistic)] for _ in self.procs]
            # ---- receive results
            state2d = self.state2d
            for arr in state2d:
                arr[:] = np.nan
            for i, conn in zip(range(nprocs), self.parent_conns):
                all_consumptions[i], all_others[i], proc_slices[i], \
                    all_changed[i], st2d = conn.recv()
                mask = ~np.isnan(st2d.scenarios)
                for j, arr in enumerate(st2d):
                    state2d[j][mask] = arr[mask]
                # add the number of previous scenarios to the one of the
                # current process `i` which doesn't see the other processes
                if i:
                    state2d.scenarios[mask] += sum(map(len,
                                                       splitted_indices[:i]))
            # choose the best scenario from all processes
            argmins = all_consumptions.argmin(axis=0)
            consumptions = all_consumptions[argmins, np.arange(nprobabilistic)]
            others = all_others[argmins]
            sl = [proc_slices[j][i] for i, j in enumerate(argmins)]
            changed_values = [all_changed[j][i] for i, j in enumerate(argmins)]
            nscenarios = np.unique(argmins).size
        if self.nprobabilistic >= 2:
            self.state, sl, changed_values = self.distribute_probabilistic(
                sl, nscenarios)
        else:
            self.state = chain(consumptions, np.squeeze(others))
            sl = sl[0]
            changed_values = changed_values[0]
            self.state2d.nscenarios[:] = 0.0
            self.state2d.nscenarios[sl] = 1.0
        if len(changed_values):
            self.data[sl] = changed_values
            # ---- send the changes to the new processes
            if self.procs is not None:
                for conn in self.parent_conns:
                    conn.send(['update', sl, changed_values, self.state])
                    conn.recv()
        # ---- store the current state and data of the model
        self.write()

    def distribute_probabilistic(self, slices, nscenarios):
        """
        Redistribute the population increase to the best scenarios

        This method distributes the population changes to the cells that have
        been computed as the best scenarios. It takes the input of the
        :meth:`best_scenario` method

        Parameters
        ----------
        slices: list
            The slicers of the best scenarios"""
        change_per_cell = self.total_change / len(slices)
        scenario_fract = 1. / len(slices)
        left_over = 0
        arr = self.data.copy()
        changed_cells = np.zeros_like(arr, dtype=np.bool)
        nscenarios_2d = self.state2d.nscenarios
        nscenarios_2d[:] = 0
        for sl in slices:
            arr[sl] = self.value_update(arr[sl], sl, change_per_cell)
            changed_cells[sl] = True
            left_over += self._left_over
            nscenarios_2d[sl] += scenario_fract
        if left_over:  # redistribute the left_over to the other cells
            for sl in slices:
                self.value_update(arr[sl], sl, left_over)
                left_over = self._left_over
                if not left_over:
                    break
        en_info = en.energy_consumption(
            arr, self.x, self.y, dist0=self.dist,
            increase=arr[changed_cells] - self.data[changed_cells],
            slicer=changed_cells,
            indices=np.arange(len(arr))[changed_cells], weights=self.weights)
        return (
            # the state
            Output(en_info[0].mean(), *en_info[1:],
                   cons_det=en._calculate_en(en_info[1:]),
                   cons_std=en_info[0].std(),
                   left_over=self._left_over, nscenarios=nscenarios),
            # the slicer
            changed_cells,
            # the changed cells
            arr[changed_cells])

    # -------------------------------------------------------------------------
    # --------------------------- First step part -----------------------------
    # ---------------------------- Initialization -----------------------------
    # -------------------------------------------------------------------------

    @property
    def init_step_methods(self):
        """Mapping from *update_method* name to the corresponding init function

        This property defines the init_step methods. Those methods are called
        at the beginning of each step on the main processor (I/O-processor).
        Each init_step method must accept no arguments and return a tuple with

        - an :attr:`slice` object or boolean array containing the information
          where the data changed
        - the indices of the cells in :attr:`data` that changed"""
        return {'categorical': _nothing,
                'random': _nothing,
                'forced': self.subtract_random}

    def subtract_random(self):
        """Subtract the people moving during this timestep"""
        movement = self.movement
        if movement == 0.0:
            return None, None
        data = self.data
        changed = 0.0
        sl = np.zeros_like(data, dtype=np.bool)
        for i in np.random.permutation(np.arange(data.size, dtype=int)):
            cell = data[i]
            if not np.isnan(cell):
                cell_change = cell * np.random.rand()
                data[i] -= cell_change
                sl[i] = True
                changed += cell_change
                if changed >= movement:
                    break
        return sl, np.arange(len(data), dtype=np.int64)[sl]

    # -------------------------------------------------------------------------
    # --------------------------- Second step part ----------------------------
    # ------------------------------- Selection -------------------------------
    # -------------------------------------------------------------------------

    @property
    def selection_methods(self):
        """Mapping from *selection_method* name to the corresponding function

        This property defines the selection methods. Those methods are called
        at the beginning of each step on the main processor (I/O-processor).
        Each selection_step method must accept no arguments and return a tuple
        with

        - an :attr:`slice` object or boolean array containing the information
          where the data changed. Alternatively it can be a list of ``None``
          and those slicing objects will be computed from the second argument
        - a 2D list of dtype integer containing the indices of the cells that
          for changed for the corresponding scenario"""
        return {'consecutive': self.consecutive_selection,
                'random': self.random_selection}

    def consecutive_selection(self):
        total_cells = self.data.size
        ncells = self.ncells
        if self.data2modify is None:
            slices = [slice(i, i + ncells) for i in range(
                0, self.data.size, ncells)]
            indices = [np.arange(i, min(i + ncells, total_cells))
                       for i in range(0, total_cells, ncells)]
        else:
            arr = self.data2modify
            indices = [
                arr[i:i + ncells] for i in range(0, arr.size, ncells)]
            slices = [None] * len(indices)
        return slices, indices

    def random_selection(self):
        if self.data2modify is None:
            indices = np.random.permutation(self.data.size)
        else:
            indices = np.random.permutation(self.data2modify)
        ncells = self.ncells
        indices = [
            indices[i:i + ncells] for i in range(0, indices.size, ncells)]
        return [None] * len(indices), indices

    # -------------------------------------------------------------------------
    # ---------------------------- Third step part ----------------------------
    # -------------------------------- Update ---------------------------------
    # -------------------------------------------------------------------------

    @property
    def update_methods(self):
        """Mapping from *update_method* name to the corresponding function

        This property defines the update methods. Each update method must
        accept and return a 1D numpy array of dtype float64 containing a view
        of the :attr:`data` attribute. The data must be modified in place!"""
        return {'categorical': self.categorical_update,
                'random': self.randomized_update,
                'forced': self.value_update}

    def value_update(self, cell_values, slicer, remaining=None):
        """Change the cells by using the forcing

        This update method changes the given cells based upon the
        :attr:`movement` information and the :attr:`population_change`
        information from the :attr:`forcing` dataset"""
        if remaining is None:
            remaining = self.total_change
        max_pop = self.max_pop[slicer]
        n = cell_values.size
        popsum = cell_values.sum()
        final_pop = max(popsum + remaining, 0)
        if remaining < 0 or final_pop <= max_pop.sum() * n:
            if n == 1:
                cell_values += remaining
                if cell_values[0] < 0:
                    remaining = cell_values[0]
                    cell_values[:] = 0
                else:
                    remaining = 0
            else:
                for i, cell in enumerate(cell_values):
                    # the maximum the current population inside the cell and
                    # otherwise the corresponding fraction at the popsum
                    change = max(-cell_values[i], min(
                        0 if popsum == 0 else (
                            remaining * (popsum - cell) / (
                                popsum * (n - 1))),
                        max_pop[i] - cell))
                    remaining -= change
                    cell_values[i] += change
        self._left_over = remaining
        return cell_values

    def categorical_update(self, cell_values, slicer):
        """Change the values through an update to the next category

        This method increases the population by updating the cells to the next
        (possible) category"""
        mask = ~np.isnan(cell_values)
        categories = self.categories
        max_pop = self.max_pop[slicer]
        indices = categories.searchsorted(cell_values[mask]) + 1
        cell_values[mask] += (
            categories[indices.clip(0, len(categories) - 1)] -
            categories[(indices-1).clip(0, len(categories) - 1)])
        cell_values[mask] = np.c_[cell_values[mask], max_pop[mask]].min(axis=1)
        return cell_values

    def randomized_update(self, cell_values, slicer):
        """Change the values through an update to a number within the next
        category

        This method increases the population by updating the cells to a random
        value within the next (possible) category"""
        def get_random(i):
            return np.random.randint(categories[i], categories[i + 1])
        categories = self.categories
        max_pop = self.max_pop[slicer]
        mask = ~np.isnan(cell_values)
        indices = categories.searchsorted(cell_values[mask]) + 1
        cell_values[mask] = list(map(get_random, indices.clip(
            0, len(categories) - 2)))
        cell_values[mask] = np.c_[cell_values[mask], max_pop[mask]].min(axis=1)
        return cell_values

    # -------------------------------------------------------------------------
    # ------------------------- Parallel Processing ---------------------------
    # -------- Necessary parts for letting the model run in parallel ----------
    # -------------------------------------------------------------------------

    #: :class:`multiprocessing.Process`. The processes of this model
    procs = None

    @property
    def nprocs(self):
        """:class:`int`. The number of processes started for this model"""
        return len(self.procs or [])

    def start_processes(self, nprocs):
        """Start `nprocs` processes for the model"""
        import multiprocessing as mp
        logger.debug('Start %i processes', nprocs)
        parent_conns = []
        child_conns = []
        for i in range(nprocs):
            parent_conn, child_conn = mp.Pipe()
            parent_conns.append(parent_conn)
            child_conns.append(child_conn)
        procs = [
            mp.Process(target=self, args=(conn, )) for conn in child_conns]
        for proc in procs:
            proc.daemon = True
            proc.start()
        self.procs = procs
        self.parent_conns = parent_conns

    def stop_processes(self):
        """Stop the processes for the model"""
        for conn in self.parent_conns:
            conn.send(None)

    def __call__(self, conn):
        logger.debug('Enter loop with %s and %s', self.consumption,
                     self.dist)
        received = True
        try:
            while received:
                received = conn.recv()
                if received:
                    try:
                        conn.send(self._step_methods[received[0]](
                            *received[1:]))
                    except Exception as e:
                        conn.close()
                        raise e
                else:
                    logger.debug('Stopping processes with %s and %s',
                                 self.consumption, self.dist)
        except EOFError:  # the connection has been closed
            pass

    @property
    def _step_methods(self):
        """The methods that are called during one :meth:`step` to synchronize
        and manage the different processes"""
        return {'best': self.best_scenario,
                'update': self.sync_state}

    def sync_state(self, sl, cell_values, state=None, weights=None):
        """
        Synchronize the :attr:`data` attribute between the processes

        This method is called to synchronize the states of the model in the
        different processes

        Parameters
        ----------
        sl: :class:`slice` or boolean np.ndarray
            The slicer that can be used to create a view on the :attr:`data`
        cell_values: np.ndarray of dtype float
            The values of the cells described by `sl`
        state: Output
            The state of the model
        """
        if sl is not None:
            self.data[sl] = cell_values
        if state is not None:
            self.state = state
        if weights is not None:
            self.weights = weights
        return

    def __reduce__(self):
        return self.__class__, (
            self.data, self.x, self.y, self.selection_method,
            self.update_model, self.ncells, self.categories,
            self.state, self.forcing, self.nprobabilistic, self.max_pop), dict(
                total_step=self.total_step, data2modify=self.data2modify,
                weights=self.weights)

    # -------------------------------------------------------------------------
    # ------------------------------- I/O Part --------------------------------
    # ----- Necessary parts for Input/Output (only called on main process) ----
    # -------------------------------------------------------------------------

    #: :class:`xarray.Dataset`. The output dataset
    _dso = None

    #: Flag that is True if the data was written to a file during the last
    #: step
    output_written = False

    #: :class:`int`. The step (corresponding to :attr:`current_step`) when the
    #: next output is written
    _next_ostep = None

    def allocate_output(self, da, dsi, steps):
        """Create the dataset for the output

        Parameters
        ----------
        da: psyplot.data.IneractiveArray
            The input data for the model
        dsi: xarray.Dataset
            The dataset `data` belongs to. If None, the
            :attr:`psyplot.data.InteractiveArray.base` attribute is used
        steps: int
            The number of steps
        """
        def empty_2d():
            ret = np.empty_like(odata)
            return ret
        logger.debug('Allocating new output dataset...')
        dso = da.to_dataset()[list(da.coords)]
        forcing = self.forcing
        if forcing is not None:
            self._tname = tname = forcing.movement.dims[0]
            current_step = self.total_step + 1
            time = forcing.coords[tname][current_step:current_step + steps]
            steps = len(time)
        else:
            tname = None
        if da.ndim == 2:
            odata = np.zeros([steps] + list(da.shape[-2:]))
            if tname is None:
                self._tname = tname = 'time'
                time = xr.Variable(('time', ), range(1, steps + 1))
            dims = [tname] + list(da.dims)
        else:
            odata = np.zeros([steps] + list(da.shape[-2:]))
            dims = da.dims
            if tname is None:
                last_time = dso[dims[0]][-1]
                self._tname = tname = last_time.name
                time = xr.concat([last_time + i for i in range(1, steps + 1)],
                                 dim=last_time.name)
            dso.drop(tname)
        dso[tname] = time
        dso[da.name] = vout = xr.Variable(dims, odata, attrs=da.attrs)

        # create other output fields
        vars2d = []
        for key, meta in fields.items():
            if key in dsi.variables:
                meta = dsi.variables[key].attrs
            dso[key] = xr.Variable((tname, ), np.zeros(steps), meta)
        for key, meta in fields2D.items():
            if key in dsi.variables:
                meta = dsi.variables[key].attrs
            vars2d.append(xr.Variable(dims, empty_2d(), meta))
            dso[key + '_2d'] = vars2d[-1]
        # create weights output field
        dso['weights'] = v_weights = xr.Variable(
            (tname, 'en_variables', 'probabilistic'),
            np.zeros((len(time), len(self.weights), self.nprobabilistic or 1)),
            {'long_name': 'Used energy consumption weights'})
        dso['en_variables'] = xr.Variable(
            ('en_variables', ), np.array(self.weights._fields),
            {'long_name': 'Variables to calculate the energy consumption'})

        self._dso = dso
        self._data_out = vout.values
        self._state2d_out = self.state2d.__class__(*[v.values for v in vars2d])
        self._weights_out = v_weights.values
        self._vname = da.name

    def write(self):
        """Write the current state to the output dataset"""
        dso = self._dso
        if dso is not None:
            i = self.current_step
            mask = self.mask
            self._data_out[i, mask] = self.data
            self._data_out[i, ~mask] = np.nan
            for key, val in six.iteritems(self.state_dict):
                dso[key][i] = val
            for j, arr in enumerate(self.state2d):
                self._state2d_out[j][i, mask] = arr
            self._weights_out[i][:, :] = self.weights
            if self._next_ostep is not None and i == self._next_ostep - 1:
                self.write_output()

    def write_output(self, complete=True):
        """Write the data to the next netCDF file

        Parameters
        ----------
        complete: bool
            If True, write the complete dataset, otherwise only until the
            current step"""
        ofile = next(self._ofiles)
        logger.info('Saving to %s...', ofile)
        if not complete:
            dso = self._dso.isel(**{
                self._tname: slice(0, self.current_step + 1)})
        else:
            dso = self._dso
        psyd.to_netcdf(dso, ofile)
        self._next_ostep = next(self._osteps)
        self.allocate_output(dso[self._vname], dsi=dso, steps=self._next_ostep)
        self.current_step = -1
        self.output_written = ofile

    docstrings.keep_params('PopulationModel.initialize_model.parameters',
                           'data', 'dsi')

    @classmethod
    @docstrings.dedent
    def get_input_ds(cls, data, dsi, **kwargs):
        """
        Return the input dataset which can be concatenated with the output

        Parameters
        ----------
        %(PopulationModel.initialize_model.parameters.data|dsi)s

        Returns
        -------
        xr.Dataset
            The modified `dsi`
        """
        # to make sure, it doesn't complain about the categories, we set a
        # value for max_pop
        kwargs['max_pop'] = kwargs.get('max_pop') or 15000
        model = cls.from_da(
            data, ofiles=[None], osteps=[2], dsi=dsi, **kwargs)
        model.current_step += 1
        model.write()
        tname = model._tname
        dsi = model._dso.isel(**{tname: slice(0, 1)}).copy()
        # estimate the first time step from the difference inside the model
        dsi[tname].values -= (
            model._dso[tname][1] - model._dso[tname][0]).values
        return dsi

    # -------------------------------------------------------------------------
    # ------------------------------ Miscallaneous ----------------------------
    # -------------------------------------------------------------------------

    @classmethod
    def _bool_or_slice(cls, arr, slicer, indices):
        """Transform the given indices to a slice object or boolean array

        Convenience method such that we do not have to give share large boolean
        arrays with other processes

        Parameters
        ----------
        arr: np.ndarray
            The data whose shape to use
        slicer: :class:`slice` object, boolean numpy.ndarray or None
            If None, a boolean array will be created using `indices and `data`.
            Otherwise, the given `slicer` is returned
        indices: np.ndarray of dtype int
            The indices of the cells that changed in `arr`

        Returns
        -------
        :class:`slice` object, boolean numpy.ndarray
            The `slicer` that can be used to create a view on `arr` for the
            given `indices`"""
        if slicer is None:
            arr_bool = np.zeros(arr.shape, dtype=bool)
            for i in indices:
                arr_bool[i] = True
            return arr_bool
        return slicer



"""Main module of the iucm package

This module defines the :class:`IUCMOrganizer` class that is used to create
a command line parser and to manage the configuration of the experiments
"""
from __future__ import print_function, division
import os
import copy
import os.path as osp
import six
import datetime as dt
from itertools import cycle
from argparse import Namespace
import logging
from model_organization import OrderedDict
from psyplot.config.rcsetup import safe_list
import iucm.utils as utils
from iucm.utils import docstrings
from model_organization import ModelOrganizer


class IUCMOrganizer(ModelOrganizer):
    """
    A class for organizing a model

    This class is indended to have hold the basic functions for organizing a
    model. You can subclass the functions ``setup, init`` to fit to your model.
    When using the model from the command line, you can also use the
    :meth:`setup_parser` method to create the argument parsers"""

    commands = ModelOrganizer.commands

    commands.insert(commands.index('archive'), 'preproc')
    commands.insert(commands.index('archive'), 'run')
    commands.insert(commands.index('archive'), 'postproc')

    paths = ModelOrganizer.paths + ['indir']

    name = 'iucm'

    # -------------------------------------------------------------------------
    # -------------------------- Preprocessing --------------------------------
    # -------------- Preprocessing functions for the experiment ---------------
    # -------------------------------------------------------------------------

    @property
    def preproc_funcs(self):
        """A mapping from preproc commands to the corresponding function"""
        return {'forcing': self.preproc_forcing,
                'mask': self.preproc_mask,
                }

    @docstrings.dedent
    def preproc(self, **kwargs):
        """
        Preprocess the data

        Parameters
        ----------
        ``**kwargs``
            Any keyword from the :attr:`preproc` attribute with kws for the
            corresponding function, or any keyword for the :meth:`main` method
        """
        funcs = self.preproc_funcs
        sp_kws = {key: kwargs.pop(key) for key in set(kwargs).intersection(
            funcs)}
        self.app_main(**kwargs)
        exp_config = self.fix_paths(self.exp_config)
        outdir = exp_config.setdefault('indir', osp.join(
            exp_config['expdir'], 'input'))
        if not osp.exists(outdir):
            os.makedirs(outdir)

        preproc_config = exp_config.setdefault('preproc', OrderedDict())

        for key, val in sp_kws.items():
            if isinstance(val, Namespace):
                val = vars(val)
            info = funcs[key](**val)
            if info:
                preproc_config[key] = info

    def _modify_preproc(self, parser):
        self._modify_app_main(parser)
        sps = parser.add_subparsers(title='Preprocessing tasks', chain=True)

        # forcing
        sp = sps.add_parser(
            'forcing',
            help='Create a forcing file from a predescribed population path')
        sp.setup_args(self.preproc_forcing)
        sp.update_arg('development_file', short='df')
        sp.update_arg('output', short='o')
        sp.update_arg('date_cols', short='t', metavar='col1,col2,col3,...',
                      type=lambda s: s.split(','))
        sp.update_arg('steps', type=int)
        sp.update_arg('movement', short='m', type=float)
        sp.update_arg('trans_size', short='trans', type=float)
        sp.update_arg('population_col', short='p')
        sp.update_arg('no_date_parsing', short='nd')

        # max_pop
        sp = parser.setup_subparser(self.preproc_mask, name='mask',
                                    return_parser=True)
        sp.update_arg('ifile', short='i')
        sp.update_arg('method', short='m')
        sp.unfinished_arguments['method']['help'] = (
            'Default: %(default)s. ' +
            sp.unfinished_arguments['method']['help'])
        sp.update_arg('hr_res', short='r')
        sp.update_arg('max_pop', short='max')
        sp.update_arg('vname', short='v')

    @docstrings.dedent
    def preproc_forcing(self, development_file=None, output=None,
                        date_cols=None, steps=1, movement=0, trans_size=0,
                        population_col=None, no_date_parsing=False):
        """
        Create a forcing file from a predescribed population path

        Parameters
        ----------
        development_file: str
            The path to a csv file containing (at least) one column with the
            projected population development
        output: str
            The name of the ouput forcing netCDF file. By default:
            ``'<expdir>/input/forcing.nc'``
        date_cols: list of str
            The names of the date columns in the `development_file` that shall
            be used to generate the date-time information. If not given, the
            date will simply be a range from 1 to `steps` times the length of
            the projected population development from `development_file`
        steps: int
            The numbers of model steps between on step of the projected
            development in `development_file`
        movement: float
            The people moving randomly during on model step
        trans_size: float
            The forced size of the transformation additionally to the
            development from `development_file`
        population_col: str
            The name of the column with population data. If not given, the last
            one is used
        no_date_parsing: bool
            If True, then `date_cols` is simply interpreted as an index and no
            date-time information is parsed
        """
        import numpy as np
        import pandas as pd
        import xarray as xr
        logger = self.logger
        logger.info('Creating forcing data...')

        idx_name = None
        if development_file is not None:
            kws = {}
            if not no_date_parsing and date_cols is not None:
                kws['parse_dates'] = {'time': date_cols}
                idx_name = 'time'
            elif date_cols is not None:
                idx_name = safe_list(date_cols)[0]
            df = pd.read_csv(development_file, **kws)
            if idx_name is not None:
                df.set_index(idx_name, inplace=True)
        else:
            df = pd.DataFrame(np.zeros((2, 1)),
                              columns=[population_col or 'population'])

        if population_col is None:
            population_col = df.columns[-1]
        steps_between = steps if not trans_size else steps * 2
        total_steps = steps_between * (len(df) - 1) + 1
        if idx_name:
            idx = np.zeros(total_steps)
            idx[:] = np.nan
            idx[::steps_between] = df.index.values
            idx = pd.Series(idx, name=idx_name)
            idx = pd.to_datetime(idx.interpolate())
        else:
            idx = None
        var = np.zeros(total_steps)
        var[:] = np.nan
        var[::steps_between] = df[population_col].values
        change = pd.Series(var, name='change', index=idx)
        change.interpolate(inplace=True)
        # only keep the population change
        change.values[1:] = change.values[1:] - change.values[:-1]
        # drop initial state
        change.drop(change.index.values[0], inplace=True)
        if trans_size:
            # only make population development every second ("add") step
            change.values[1::2] += change.values[:-1:2] + trans_size
            change.values[::2] = -trans_size
        forcing_df = change.to_frame()
        forcing_df['movement'] = movement
        if not forcing_df.index.name:
            forcing_df.index.name = 'step'
        ds = xr.Dataset.from_dataframe(forcing_df)
        ds.variables['change'].attrs.update(
            {'long_name': "population change", 'units': 'inhabitants'})
        ds.variables['movement'].attrs.update(
            {'long_name': "randomly moving population",
             'units': 'inhabitants'})
        if output is None:
            output = output = self.exp_config.get(
                'forcing', osp.join(self.exp_config['indir'], 'forcing.nc'))
        logger.debug('Saving output to %s', output)
        ds.to_netcdf(output)
        self.exp_config['forcing'] = output

        config = OrderedDict([
            ('development_file', development_file),
            ('development_steps', list(range(
                    0, total_steps, steps_between))[1:]),
            ('trans_size', trans_size),
            ('total_steps', total_steps - 1),
            ('movement', movement)])
        if logger.isEnabledFor(logging.DEBUG):
            for key, val in config.items():
                logger.debug('    %s: %s', key, val)

        return config

    @docstrings.dedent
    def preproc_mask(self, shapefile, method='max-pop', ifile=None, vname=None,
                     overwrite=False, max_pop=None, hr_res=100):
        """
        Mask grid cells based on a shape file

        This command calculates the maximum population for the model based on
        a masking shape file. The given shape file is rasterized at high
        resolution (by default, 100 times the resolution of the input file) and
        the fraction for each grid cell that is covered by that shape file is
        calculated.

        Parameters
        ----------
        shapefile: str
            The path to a shapefile
        method: {'max-pop', 'mask', 'ignore'}
            Determines how to handle the given shapes.

            max-pop
                The maximum population per grid cell is lowered by the fraction
                of the cell that is covered by the given shape. This will
                adjust the `max_pop` variable in the input file `ifile`
            mask
                The population of the grid cells in the input data that are
                touched by the given shapes will be kept constant during the
                simulation. This will adjust the `mask` variable in the input
                file `ifile`
            ignore
                The grid cells in the input data that are touched by the
                given grid cells are put to NaN and their population is not
                considered during the simulation. This will adjust the input
                population data (i.e. variable `vname`) directly
        ifile: str
            The path of the input file. If not specified, the value of the
            configuration is used
        vname: str
            The variable name to use. If not specified and only one variable
            exists in the dataset, this one is used. Otherwise, the
            ``'run.vname'`` key in the experiment configuration is used
        overwrite: bool
            If True and the target variable exists already in the input file
            `ifile` (and method is not 'ignore'), this variable is overwritten
        max_pop: float
            The maximum population. If not specified, the value in the
            configuration is used (only necessary if ``method=='max-pop'``)
        hr_res: int
            The resolution of the high resolution file, relative to the
            resolution of `ifile` (only necessary if ``method=='max-pop'``)

        Notes
        -----
        Note that the shapefile and the input file have to be defined on the
        same coordinate system! This function is not super efficient, for large
        data files we recommend using gdal_rasterize_ and gdalwarp_.

        .. _gdal_rasterize: http://www.gdal.org/gdal_rasterize.html
        .. _gdalwarp: http://www.gdal.org/gdalwarp.html"""
        from shapefile import Reader
        import matplotlib.path as mplp
        import psyplot.data as psyd
        import numpy as np
        import netCDF4 as nc
        exp_config = self.exp_config
        run_config = exp_config.setdefault('run', OrderedDict())
        if max_pop is None:
            max_pop = run_config.get('max_pop')
        if ifile is None:
            ifile = exp_config.get('input')
        if method == 'max-pop' and max_pop is None:
            raise ValueError(
                "The maximum population per grid cell has not yet been "
                "specified! See `iucm preproc mask --help`")
        if ifile is None:
            raise ValueError(
                "The input file needs to be specified! "
                "See `iucm preproc mask --help`")
        dsi = psyd.open_dataset(ifile, decode_times=False)
        vname = vname or run_config.get('vname') or self.get_population_vname(
            dsi)

        reader = Reader(shapefile)
        paths = list(mplp.Path(np.asarray(shape.points)[..., :2])
                     for shape in reader.shapes())

        x = dsi.psy[vname].psy.get_coord('x').values
        y = dsi.psy[vname].psy.get_coord('y').values
        dims = dsi.psy[vname].dims[-2:]
        ny, nx = dsi[vname].shape[-2:]

        dsi.close()
        del dsi

        if method == 'max-pop':
            hr_x = np.linspace(y[0], x[-1], len(x) * hr_res)
            hr_y = np.linspace(y[0], y[-1], len(y) * hr_res)

            X, Y = np.meshgrid(hr_x, hr_y)
        else:
            X, Y = np.meshgrid(x, y)

        points = np.array((X.flatten(), Y.flatten())).T
        del X, Y

        mask = np.zeros(points.shape[:-1], dtype=bool)
        for p in paths:
            mask |= p.contains_points(points)

        if method == 'max-pop':
            arr = mask.reshape(ny, hr_res, nx, -1).astype(float).mean(
                axis=(1, 3))
        else:
            mask = mask.reshape((ny, nx))

        nco = nc.Dataset(ifile, 'a')
        vnames = {'max-pop': 'max_pop', 'mask': 'mask', 'ignore': vname}
        target_vname = vnames[method]
        create = target_vname not in nco.variables
        if create:
            var = nco.createVariable(
                target_vname, int if method == 'mask' else float, dims)
            if method == 'max-pop':
                attrs = {'long_name': 'Maximum population',
                         'units': 'inhabitants'}
            else:
                attrs = {'long_name': 'Cells with constant population',
                         'comments': ('Cells with a non-zero value are not '
                                      'modified by the IUCM simulation.')}
            var.setncatts(attrs)
        else:
            var = nco.variables[target_vname]
        if method == 'max-pop' and (create or overwrite):
            var[:] = max_pop
        elif method == 'mask' and (create or overwrite):
            var[:] = 0
        if method == 'max-pop':
            var[:] *= (1 - arr)
        elif method == 'mask':
            var[:] = (mask | var[:].astype(bool)).astype(int)
        else:
            arr = var[:]
            arr[mask] = np.nan
            var[:] = arr

        nco.close()

    # -------------------------------------------------------------------------
    # ------------------------------- Run -------------------------------------
    # --------------------------- Run the experiment --------------------------
    # -------------------------------------------------------------------------

    @docstrings.dedent
    def run(self, ifile=None, forcing=None, vname=None, steps=50,
            selection_method=None, update_method=None, ncells=None,
            categories=None, use_pctls=False, no_restart=False,
            output_step=None, seed=None, stop_en_change=None,
            agg_stop_steps=100, agg_steps=1, probabilistic=None,
            max_pop=None, coord_transform=None, copy_from=None, **kwargs):
        """
        Run the model for the given experiment

        Parameters
        ----------
        ifile: str
            The input file. If not specified, the `input` key in the
            experiment configuration is used
        forcing: str
            The forcing file (necessary if ``update_method=='forced'``). If not
            specified, the `forcing` key in the experiment configuration is
            used
        vname: str
            The variable name to use. If not specified and only one variable
            exists in the dataset, this one is used. Otherwise, the
            ``'run.vname'`` key in the experiment configuration is used
        steps: int
            The number of time steps
        selection_method: { 'consecutive' | 'random' }
            The name of the method on how the data is selected.
            The default is consecutive.
            Possible selection methods are

            consecutive:
                Always the `ncells` consecutive cells are selected.
            random:
                `ncells` random cells in the field are updated.
        update_method: { 'categorical' | 'random' | 'forced' }
            The name of the update method on how the selected cells (see
            `selection_method` are updated). The default is categorical.
            Possible update methods are

            categorical:
                The selected cells are updated to the lower level of the next
                category.
            random:
                The selected cells are updated to a randum number within the
                next category.
            forced:
                A forcing file is used (see the `forcing` parameter).
        ncells: int
            The number of cells that shall be changed during 1 step. The
            default value is 4
        categories: list of floats
            The values for the categories to use within the models
        use_pctls: bool
            If True, interprete `categories` as percentiles instead of real
            population density
        no_restart: bool
            If True, and the run has already been conducted, restart it.
            Otherwise the previous run is continued
        output_step: int
            Make an output every `output_step`. If None, only the final result
            is written to the output file
        seed: int
            The random seed for numpy to use. Specify this parameter for the
            experiment to guarantee reproducability
        stop_en_change: float
            The minimum of required relative energy consumption change. If the
            mean relative energy consumption change over the last
            `agg_stop_steps` steps is below this number, the run is stopped
        agg_stop_steps: int
            The number of steps to aggregate over when calculating the mean
            relative energy consumption change. Does not have an effect if
            `stop_en_change` is None
        agg_steps: int
            Use only every `agg_steps` energy consumption for the aggregation
            when checking the `stop_en_change` criteria
        probabilistic: int
            The number of probabilistic scenarios. For each scenario the energy
            consumption is calculated and the final population is distributed
            to the cells with the ideal energy consumption. Set this to 0 to
            only use the weights by [LeNechet2012]_. If this option is None,
            the value will be taken from the configuration with a default of 0
            (i.e. no probabilistic run).
        max_pop: float
            The maximum population for each cell. If None, the last
            value in `categories` will be used or what is stored in the
            experiment configuration
        coord_transform: float
            The transformation factor to transform the coordinate values into
            kilometres
        copy_from: str
            If not None, copy the run settings from the other given experiment
        ``**kwargs``
            Any other keyword argument that is passed to the :meth:`main`
            method"""
        import numpy as np
        import psyplot.data as psyd
        from iucm.model import PopulationModel

        def get_files(imin, imax):
            if output_step is None:
                return [osp.join(outdir, exp + '_%i-%i.nc' % (imin, imax))]
            else:
                return [osp.join(outdir, exp + '_%i-%i.nc' % (i, min(
                            i + output_step - 1, imax)))
                        for i in range(imin, imax + 1, output_step)]

        def log_state():
            for key, val in six.iteritems(model.state._asdict()):
                logger.debug('    %s: %s', key, val)

        self.app_main(**kwargs)
        exp = self.experiment
        logger = self.logger
        debug = logger.isEnabledFor(logging.DEBUG)
        exp_config = self.fix_paths(self.exp_config)
        global_conf = self.config.global_config
        run_config = exp_config.setdefault('run', OrderedDict())
        if copy_from is not None:
            run_config.update(
                self.config.experiments[copy_from].get('run', {}))
            try:
                exp_config['forcing'] = self.config.experiments[copy_from][
                    'forcing']
            except KeyError:
                pass
        logger.debug('Run experiment...')

        seed = seed or exp_config['run'].get('seed')
        if seed is not None:
            np.random.seed(int(seed))
            run_config['seed'] = seed

        # ----------------------- Model Configuration -------------------------
        defaults = {'ncells': 4, 'max_pop': None,
                    'selection_method': 'consecutive',
                    'update_method': 'forced',
                    'probabilistic': 0,
                    'categories': None,
                    'coord_transform': 1.,
                    }
        run_settings = {'ncells': ncells,
                        'selection_method': selection_method,
                        'update_method': update_method,
                        'categories': categories,
                        'probabilistic': probabilistic,
                        'max_pop': max_pop,
                        'coord_transform': coord_transform,
                        }
        for key, val in run_settings.items():
            run_settings[key] = val = val if val is not None else \
                run_config.get(key, defaults[key])
            run_config[key] = val
        # get the actual steps value. The run_config will be updated below,
        # depending on whether we continue a run or not

        # ------------------------- output configuration ----------------------
        continued = 'run' in exp_config['timestamps']
        outdir = exp_config.setdefault('outdir', osp.join(
            exp_config['expdir'], 'outdata'))
        if not osp.exists(outdir):
            os.makedirs(outdir)
        if no_restart or not continued:
            for f in exp_config.get('outdata', []):
                if osp.exists(f):
                    os.remove(f)
            exp_config['outdata'] = []
            outfiles = get_files(1, steps)
            run_config['steps'] = 0
            run_config['step_date'] = OrderedDict()
            old_steps = 0
        elif continued:
            old_steps = run_config['steps']
            outfiles = get_files(old_steps + 1, old_steps + steps)
            run_config['steps'] += steps
            ifile = exp_config['outdata'][-1]
            run_config.setdefault('step_date', OrderedDict())
            if categories is None:  # percentiles have already been computed
                use_pctls = False
        else:
            raise ValueError("Cannot restart the experiment because no run "
                             "was ever invoked")
        logger.debug('    Output files: %s', outfiles)
        if output_step is None:
            output_step = steps
        outfiles = iter(outfiles)
        output_step = iter(cycle(safe_list(output_step)))

        # -------------------------- input file -------------------------------
        indir = exp_config.get('indir',
                               osp.join(exp_config['expdir'], 'input'))
        if not osp.exists(indir):
            os.makedirs(indir)
        if ifile is None:
            ifile = exp_config.get('input', self.project_config.get(
                'input',  self.global_config.get('input')))
        if ifile is None:
            raise ValueError("No input file specified!")
        if not continued:
            ifile_conf = osp.join(indir, 'input.nc')
            exists = osp.exists(ifile_conf)
            if exists and not osp.samefile(ifile_conf, ifile):
                os.remove(ifile_conf)
                os.symlink(osp.relpath(ifile, indir), ifile_conf)
            elif not exists:
                os.symlink(osp.relpath(ifile, indir), ifile_conf)
            exp_config['input'] = osp.abspath(ifile_conf)
        logger.debug('    Input file: %s', ifile)

        # -------------------------- forcing file -----------------------------
        if forcing is None:
            forcing = exp_config.get('forcing', self.project_config.get(
                'forcing',  self.global_config.get('forcing')))
        logger.debug('    Forcing file: %s', forcing)
        if forcing is not None:
            exp_config['forcing'] = osp.abspath(forcing)
            forcing = psyd.open_dataset(forcing)

        # --------------------------- Load data -------------------------------
        dsi = psyd.open_dataset(ifile)
        vname = vname or run_config.get('vname') or self.get_population_vname(
            dsi)
        try:
            data = dsi[vname]
        except KeyError:
            vnames = set(dsi.variables) - set(dsi.coords) - {'mask', 'max_pop'}
            if vname == run_config.get('vname'):
                msg = ('The variable %s as stored in the configuration is not '
                       'found in the dataset. Possible names are {%s}. Please '
                       'use the `vname` parameter!') % (
                           vname, ', '.join(vnames))
            else:
                msg = ('The variable %s is not found in the dataset. Possible '
                       'names are {%s}.') % (vname, ', '.join(vnames))
            raise KeyError(msg)
        run_config['vname'] = data.name

        if (coord_transform is None and
                data[data.dims[-1]].attrs.get('units', None) in [
                        'm', 'meter', 'metres', 'metre', 'meters']):
            run_config['coord_transform'] = run_settings['coord_transform'] = \
                coord_transform = 1000.

        logger.debug('    Input variable: %s, shape %s', vname, data.shape)

        logger.debug('Initialize model...')
        model = PopulationModel.from_da(
            data, forcing=forcing, ofiles=outfiles, osteps=output_step,
            use_pctls=use_pctls, dsi=dsi, last_step=old_steps, **run_settings)
        if model.categories is not None:
            run_config['categories'] = model.categories.tolist()
        if debug:
            logger.debug('Initial state:')
            log_state()

        # ------------------ Start parallel processes -------------------------
        if not global_conf.get('serial'):
            import multiprocessing as mp
            nprocs = global_conf.get('nprocs', 'all')
            if nprocs == 'all':
                nprocs = mp.cpu_count()
            model.start_processes(nprocs)
        logger.debug('Start main loop')
        t = [dt.datetime.now()] * (steps + 1)
        # --------------------------- Start loop ------------------------------
        last_output = None
        if stop_en_change is not None:
            consumptions = np.zeros(steps + 1)
            consumptions[0] = model.consumption
            # to use exactly the specified number of `agg_stop_steps` steps
            # when calculating the changes, we have to subtract 1
            agg_stop_steps -= 1
        for step in range(1, steps + 1):
            logger.debug('Step %i', step)
            model.step()
            if debug:
                log_state()
                t[step] = dt.datetime.now()
            if model.output_written and not self.no_modification:
                exp_config['outdata'].append(model.output_written)
                run_config['steps'] = old_steps + step
                if last_output is not None:
                    run_config['step_date'].pop(old_steps + last_output, None)
                last_output = step
                run_config['step_date'][old_steps + last_output] = str(t[step])
                exp_config['timestamps']['run'] = str(t[step])
                copied = copy.deepcopy(self.config)
                self.rel_paths(copied.experiments[self.experiment])
                copied.save()
            if stop_en_change is not None:
                consumptions[step] = model.consumption
                arr = consumptions[step - agg_stop_steps:step + 1:agg_steps]
                any_inf = np.isinf(arr).any()
                mean_change = np.abs((arr[1:] - arr[:-1]) / arr[:-1]).mean()
                if any_inf or mean_change <= stop_en_change:
                    logger.info("Stopping after %i steps with mean change %s",
                                step, mean_change)
                    break
        try:  # save the last step
            model.write_output(False)
        except StopIteration:  # last step already saved
            pass
        else:
            exp_config['outdata'].append(model.output_written)
            run_config['steps'] = old_steps + steps
        if debug:
            diffs = np.array([(t[i+1] - t[i]).seconds for i in range(steps)])
            logger.debug('Average step time: %s seconds' % diffs.mean())
            logger.debug('Fastest step time: %s seconds' % diffs.min())
            logger.debug('Slowest step time: %s seconds' % diffs.max())

        # --------------------------- Finish loop -----------------------------
        if not global_conf.get('serial'):
            model.stop_processes()

        logger.debug('Run Done.')

    def get_population_vname(self, ds):
        vnames = list(filter(
            lambda v: ds[v].ndim >= 2,
            set(ds.variables) - set(ds.coords) - {'mask', 'max_pop'}))
        if len(vnames) == 0:
            raise ValueError("Found no variables in the input dataset!")
        if len(vnames) > 1:
            raise ValueError(
                "Found %i possible variable names. Please specify one of "
                "{%s}" % (len(vnames), ', '.join(vnames)))
        return vnames[0]

    def _modify_run(self, parser):
        parser.update_arg('ifile', short='i')
        parser.update_arg('forcing', short='f')
        parser.update_arg('vname', short='v')
        parser.update_arg('steps', short='t', type=int)
        parser.update_arg('selection_method', short='sm')
        parser.update_arg('update_method', short='um')
        parser.update_arg('ncells', short='n', type=int)
        parser.update_arg('categories', short='c', metavar='float1,float2,...',
                          type=lambda s: list(map(int, s.split(','))))
        parser.pop_key('categories', 'nargs', None)
        parser.update_arg('no_restart', short='nr')
        parser.update_arg('output_step', short='ot', type=int)
        parser.update_arg('seed', type=int)
        parser.update_arg('use_pctls', short='pctls')
        parser.update_arg('probabilistic', short='prob')
        parser.update_arg('copy_from', short='cp')
        parser.update_arg('max_pop', short='max')
        parser.update_arg('coord_transform', short='ct')

    # -------------------------------------------------------------------------
    # -------------------------- Postprocessing -------------------------------
    # ------------ Postprocessing functions for the experiment ----------------
    # -------------------------------------------------------------------------

    @property
    def postproc_funcs(self):
        """A mapping from postproc commands to the corresponding function"""
        return OrderedDict([('rolling', self.rolling_mean),
                            ('map', self.make_map),
                            ('movie', self.make_movie),
                            ('evolution', self.plot_evolution),
                            ])

    @docstrings.dedent
    def postproc(self, no_input=False, **kwargs):
        """
        Postprocess the data

        Parameters
        ----------
        no_input: bool
            If True/set, the initial input file is ignored"""
        import xarray as xr
        funcs = self.postproc_funcs
        sp_kws = {key: kwargs.pop(key) for key in set(kwargs).intersection(
            funcs)}
        self.app_main(**kwargs)
        exp_config = self.fix_paths(self.exp_config)
        outdir = exp_config.setdefault('postprocdir', osp.join(
            exp_config['expdir'], 'postproc'))
        if not osp.exists(outdir):
            os.makedirs(outdir)
        files = exp_config['outdata']
        postproc_config = exp_config.setdefault('postproc', OrderedDict())

        ds = xr.open_mfdataset(files)

        if not no_input:
            from iucm.model import PopulationModel
            dsi = xr.open_dataset(exp_config['input'])
            vname = exp_config['run']['vname']
            forcing = exp_config.get('forcing')
            if forcing is not None:
                forcing = xr.open_dataset(forcing)
            dsi = PopulationModel.get_input_ds(
                dsi[vname], dsi, forcing=forcing)
            ds = xr.concat([dsi, ds], dim=dsi[vname].dims[0])

        for key, func in funcs.items():
            if key not in sp_kws:
                continue
            val = sp_kws[key]
            if isinstance(val, Namespace):
                val = vars(val)
                val.pop('no_input', None)
            info = func(ds, **val)
            if info:
                postproc_config[key] = info

    def _modify_postproc(self, parser):
        self._modify_app_main(parser)
        parser.update_arg('no_input', short='ni')
        sps = parser.add_subparsers(title='Postprocessing tasks', chain=True)

        # rolling mean
        sp = sps.add_parser(
            'rolling',
            help='Calculate the rolling mean for the energy consumption')
        sp.setup_args(self.rolling_mean)
        sp.pop_arg('ds')
        sp.update_arg('window', short='w')
        sp.update_arg('output', short='o')
        sp.update_arg('odir', short='od')

        # map
        sp = sps.add_parser(
            'map', help='Make a 2D-plot of the population')
        sp.setup_args(self.make_map)
        sp.pop_arg('ds')
        sp.pop_arg('close')
        sp.update_arg('output', short='o')
        sp.update_arg('odir', short='od')
        sp.update_arg('project_file', short='p')
        sp.update_arg('save_project', short='save')
        sp.update_arg('simple_plot', short='simple')
        sp.update_arg('time', short='t', type=int)
        sp.update_arg('t0', type=int)

        # movie
        sp = sps.add_parser(
            'movie', help='Make a movie of the population development')
        sp.setup_args(self.make_movie)
        sp.pop_arg('ds')
        sp.pop_arg('close')
        sp.update_arg('output', short='o')
        sp.update_arg('odir', short='od')
        sp.update_arg('project_file', short='p')
        sp.update_arg('save_project', short='save')
        sp.update_arg('simple_plot', short='simple')
        sp.update_arg('t0', type=int)
        sp.update_arg('time', short='t', type=utils.str_ranges,
                      help=docstrings.dedents("""
                          The time steps to use. %(str_ranges.s_help)s"""),
                      metavar='t1[;t2[;t31,t32,[t33]]]')

        sp.add_argument('-fps', help=(
            "The number of frames per second. Default: 10"), type=int)
        sp.add_argument('-dpi', help=("The dots per inch"), type=int)
        # evolution
        sp = sps.add_parser('evolution', help=(
            'Plot the evolution of energy consumption, DIST, RSS and ENTROP'))
        sp.setup_args(self.plot_evolution)
        sp.pop_arg('ds')
        sp.pop_arg('close')
        sp.update_arg('output', short='o')
        sp.update_arg('odir', short='od')
        sp.update_arg('time', short='t', type=utils.str_ranges,
                      help=docstrings.dedents("""
                          The time steps to use. %(str_ranges.s_help)s"""),
                      metavar='t1[;t2[;t31,t32,[t33]]]')
        sp.update_arg('use_rolling', short='roll')

    @docstrings.get_sectionsf('IUCMOrganizer.rolling_mean')
    @docstrings.dedent
    def rolling_mean(self, ds, window=None, output=None, odir=None):
        """
        Calculate the rolling mean for the energy consumption

        This postprocessing function calculates the rolling mean for the
        energy consumption

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset with the *cons* and *cons_std* variables
        window: int
            Size of the moving window. This is the number of observations used
            for calculating the statistic. Each window will be a fixed size. If
            None, it will be taken from the experiment configuration with a
            default value of 50.
        output: str
            A filename where to save the output. If not given, it is not saved
            but may be later used by the :meth:`evolution` method
        odir: str
            The name of the output directory
        """
        if window is None:
            window = self.exp_config.get('postproc', {}).get(
                'rolling', {}).get('window', 50)
        self.logger.debug('Calculating rolling mean with window of %i...',
                          window)
        rolling = ds.cons.to_pandas().rolling(window)
        ds.variables['cons'][:] = rolling.mean()
        ds.variables['cons_std'][:] = rolling.std()
        if odir is None:
            odir = osp.join(self.exp_config['expdir'], 'postproc')
        if output:
            output = osp.join(odir, output)
            ds[['cons', 'cons_std']].to_netcdf(output)
            return {'output': output, 'window': window}
        return {'window': window}

    @docstrings.get_sectionsf('IUCMOrganizer.make_map')
    @docstrings.dedent
    def make_map(self, ds, output='map.pdf', odir=None, time=-1, diff=False,
                 t0=0, project_file=None, save_project=None, simple_plot=False,
                 close=True, **kwargs):
        """
        Make a movie of the post processing

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset for the plot or a list of filenames
        output: str
            The name of the output file
        odir: str
            The name of the output directory
        time: int
            The timestep to plot. By default, the last timestep is used
        diff: bool
            If True/set, visualize the difference to the `t0` (by default, the
            first step) is used, instead of the entire data
        t0: int
            If `diff` is set, the reference step for the difference
        project_file: str
            The path to a filename containing a file that can be loaded via
            the :meth:`psyplot.project.Project.load_project` method
        save_project: str
            The path to a filename where to save the psyplot project
        simple_plot: bool
            If True/set, use a non-georeferenced plot. Otherwise, we use the
            cartopy module to plot it
        close: bool
            If True, close the project at the end

        Other Parameters
        ----------------
        ``**kwargs``
            Any other keyword that is passed to the
            :meth:`psyplot.project.Project.export` method"""
        import numpy as np
        import psyplot.project as psy
        from psyplot.data import _infer_interval_breaks

        self.logger.debug('Create plot of step %s...', time)

        config = self.exp_config['postproc'].get('map', OrderedDict())

        kwargs.update({key: config[key] for key in set(config) - {
            'plot_output', 'project_output'}})

        if odir is None:
            odir = osp.join(self.exp_config['expdir'], 'postproc')

        #: name of the variable in the above netCDF files
        vname = self.exp_config['run']['vname']

        if diff:
            ds = ds.copy(True)
            arr = ds[vname].values
            mask = np.zeros_like(arr, dtype=bool)
            diff = arr[time] - arr[0]
            mask[time] = ~np.isnan(arr[time]) & (diff == 0)
            arr[mask] = 0

        if project_file is not None:
            sp = psy.Project.load_project(project_file, datasets=[ds])
        else:
            #: categories of the experiment (will also be used as colorbar
            #: ticks)
            categories = np.array([0] + self.exp_config['run']['categories'])
            #: boundaries of the colormap
            bounds = _infer_interval_breaks(categories)
            bounds[0] = 0

            # ---- make the plot
            #: psyplot project of the plot
            if simple_plot:
                sp = psy.plot.plot2d(
                    ds, name=vname, mf_mode=True, t=time,
                    bounds=bounds, cmap='afmhot_r', cticks=categories,
                    title='step %(time)s', titleprops={'ha': 'left'})
            else:
                sp = psy.plot.mapplot(
                    ds, name=vname, mf_mode=True, t=time,
                    bounds=bounds, transform='moll', cmap='afmhot_r',
                    cticks=categories, projection='moll',
                    title='step %(time)s', titleprops={'ha': 'left'},
                    xgrid=False, ygrid=False)

        #: export the plot
        if output:
            output = osp.join(odir, output)
            self.logger.debug('Saving plot to %s', output)
            sp.export(output, **kwargs)
            kwargs['plot_output'] = output

        if save_project is None:
            save_project = config.get('project_output')
        if save_project:
            save_project = osp.join(odir, save_project)
            sp.save_project(save_project, dump=False,
                            paths=[self.exp_config['outdata']])
            kwargs['project_output'] = save_project

        if close:
            sp.close(True, True)
            return kwargs
        return kwargs, sp

    docstrings.delete_params('IUCMOrganizer.make_map.parameters', 'time')

    @docstrings.dedent
    def make_movie(self, ds, output='movie.gif', odir=None, diff=False,
                   t0=None, project_file=None, save_project=None,
                   simple_plot=False, close=True, time=None, **kwargs):
        """
        Make a movie of the post processing

        Parameters
        ----------
        %(IUCMOrganizer.make_map.parameters.no_time)s
        time: list of int
            The time steps to use for the movie

        Other Parameters
        ----------------
        ``**kwargs``
            Any other keyword for the
            :class:`matplotlib.animation.FuncAnimation` class that is used to
            make the plot"""
        import numpy as np
        from matplotlib.animation import FuncAnimation

        self.logger.debug('Making movie...')

        if odir is None:
            odir = osp.join(self.exp_config['expdir'], 'postproc')

        task_names = {True: 'diff', False: 'regular'}
        task_name = task_names[diff]

        config = self.exp_config['postproc'].get('movie', OrderedDict()).get(
            task_name, OrderedDict())

        defaults = {'fps': 10}

        for key in set(config) - {'plot_output', 'project_output'}:
            if kwargs[key] is None:
                kwargs.setdefault(key, config[key])

        # update with defaults for the movie writer
        for key, val in defaults.items():
            if kwargs.get(key) is None:
                kwargs[key] = val

        #: name of the variable in the above netCDF files
        vname = self.exp_config['run']['vname']
        tname = ds[vname].dims[0]

        if diff and t0 is not None:
            ref = ds.isel(**{tname: t0})[vname].values

        orig_time = range(ds[tname].size)

        if time is not None:
            if diff and t0 is None:
                ref = ds.isel(**{tname: safe_list(time)[0]})[
                    vname].values
                t0 = 0
            if isinstance(time, slice):
                orig_time = time.indices(ds[tname].size)
            else:
                orig_time = time[:]
            ds = ds.isel(**{tname: time})
        elif diff and t0 is None:
            ref = ds.isel(**{tname: 0})[vname].values
            t0 = 0

        time = range(ds[tname].size)

        if diff:
            ds = ds.copy(True)
            arr = ds[vname].values
            for i, orig in zip(range(0, ds[tname].size), orig_time):
                if orig != t0:
                    mask = np.zeros_like(arr, dtype=bool)
                    diffs = arr[i] - ref
                    mask[i] = ~np.isnan(arr[i]) & (diffs == 0)
                    arr[mask] = 0

        _, sp = self.make_map(
            ds, odir=odir, diff=False, output=None, project_file=project_file,
            save_project=False, simple_plot=simple_plot, close=False,
            time=0)

        def update(t):
            """Update function for the movie"""
            sp.update(t=t)

        #: animation which create the movie
        ani = FuncAnimation(
            next(iter(sp.figs)), update, time, sp.draw)
        #: save the movie to a gif file
        if output:
            output = osp.join(odir, output)
            self.logger.debug('Saving movie to %s', output)
            self.logger.debug('Movie config: %s', kwargs)
            ani.save(output, **kwargs)
            kwargs['plot_output'] = output

        save_project = save_project or config.get('project_output')
        if save_project:
            save_project = osp.join(odir, save_project)
            sp.save_project(save_project, dump=False,
                            paths=[self.exp_config['outdata']])
            kwargs['project_output'] = save_project

        info = {task_name: kwargs}
        if self.exp_config['postproc'].get('movie', OrderedDict()).get(
                task_names[not diff], OrderedDict()):
            info[task_names[not diff]] = self.exp_config['postproc']['movie'][
                task_names[not diff]]

        if close:
            sp.close(True, True)
            return info

        return info, sp

    docstrings.keep_params('IUCMOrganizer.make_map.parameters', 'ds',
                           'output', 'odir')

    @docstrings.dedent
    def plot_evolution(self, ds, output='plots.pdf', odir=None, close=True,
                       time=None, use_rolling=False):
        """
        Plot the evolution of DIST, RSS, ENTROP and Energy consumption

        Parameters
        ----------
        %(IUCMOrganizer.make_map.parameters.ds|output|odir)s
        close: bool
            If True, the created figures are closed in the end
        time: list of int
            The time steps to use for the movie
        use_rolling: bool
            If True, use the rolling mean for the energy consumption

        Returns
        -------
        dict
            Information on the output"""
        import numpy as np
        import matplotlib.pyplot as plt
        import seaborn as sns
        from matplotlib.backends.backend_pdf import PdfPages
        import iucm.energy_consumption as en

        self.logger.debug('Plot timeseries...')

        if odir is None:
            odir = osp.join(self.exp_config['expdir'], 'postproc')

        if use_rolling:
            import xarray as xr
            if 'output' in self.exp_config.get('postproc', {}).get(
                    'rolling', {}):
                with xr.open_dataset(self.exp_config['postproc'][
                        'rolling']['output']) as ds2:
                    ds.variables['cons'] = ds2.variables['cons']
                    ds.variables['cons_std'] = ds2.variables['cons_std']
            else:
                self.rolling_mean(ds)

        if time is not None:
            ds = ds.isel(**{ds.dist.dims[0]: time})

        ds_plot = ds[['cons', 'rss', 'dist', 'entrop', 'cons_std']].copy(True)
        ds_plot2 = ds_plot.copy(True)

        # ----- area plot
        ds_plot['entrop'].values *= en.wENTROP
        ds_plot['rss'].values *= en.wRSS
        ds_plot['dist'].values *= en.wDIST
        summed = np.abs(ds_plot.dist) + np.abs(ds_plot.rss) + np.abs(
            ds_plot.entrop)
        for vname in ['dist', 'rss', 'entrop']:
            ds_plot[vname].values /= summed

        #: List of matplotlib figures created in this function
        figs = []

        figs.append(plt.figure())

        plt.stackplot(
            ds_plot.time.values, ds_plot.dist.values, ds_plot.entrop.values,
            labels=[ds.dist.long_name, ds.entrop.long_name], baseline='zero')
        plt.stackplot(ds_plot.time.values, ds_plot.rss.values,
                      labels=[ds.rss.long_name])

        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.2)
        plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.1), ncol=3)

        # ----- single line plot for each variable
        # --- population
        figs.append(plt.figure())
        pop_da = ds[self.exp_config['run']['vname']]
        pop_da.sum(axis=(1, 2)).plot()
        plt.ylabel('Population [inhabitants]')
        # --- energy consumption
        units = ds.cons.units
        figs.append(plt.figure())
        ds_plot.cons.plot()
        plt.fill_between(ds_plot[ds_plot.cons.dims[0]],
                         ds_plot.cons - ds_plot.cons_std,
                         ds_plot.cons + ds_plot.cons_std,
                         facecolors=plt.gca().lines[0].get_c(),
                         edgecolors='None',
                         alpha=0.5)
        plt.ylabel('%s [%s]' % (ds.cons.long_name, units))

        ds_plot2['dist'].values *= en.wDIST
        ds_plot2['rss'].values *= en.wRSS
        ds_plot2['entrop'].values *= en.wENTROP
        for name in ['dist', 'rss', 'entrop']:
            figs.append(plt.figure())
            ds_plot2[name].plot()
            plt.ylabel('weighted %s [%s]' % (ds_plot2[name].long_name, units))

        ret = {}

        if output:
            output = osp.join(odir, output)
            self.logger.debug('Saving plots to %s', output)
            with PdfPages(output) as pdf:
                for fig in figs:
                    pdf.savefig(fig, bbox_inches='tight')
            ret['plot_output'] = output
        if close:
            for fig in figs:
                plt.close(fig.number)
        return ret


def _get_parser():
    """Function returning the iucm parser, necessary for sphinx documentation
    """
    return IUCMOrganizer.get_parser()


def main():
    """Call the :meth:`~model_organization.ModelOrganizer.main` method of the
    :class:`IUCMOrganizer` class"""
    return IUCMOrganizer.main()


if __name__ == '__main__':
    main()



"""General utilities for the iucm package"""
import six
from itertools import chain
from docrep import DocstringProcessor

docstrings = DocstringProcessor()


docstrings.params['str_ranges.s_help'] = """
    A semicolon (``';'``) separated string. A single value in this string
    represents one number, ranges can also be used via a separation by
    comma (``','``). Hence, ``'2009;2012,2015'`` will be
    converted to ``[2009,2012, 2013, 2014]`` and ``2009;2012,2015,2`` to
    ``[2009, 2012, 2015]``"""


@docstrings.dedent
def str_ranges(s):
    """
    Convert a string of comma separated values to an iterable

    Parameters
    ----------
    s: str%(str_ranges.s_help)s

    Returns
    -------
    list
        The values in s converted to a list"""
    def get_numbers(s):
        nums = list(map(int, s.split(',')))
        if len(nums) == 1:
            return nums
        else:
            import numpy as np
            return np.arange(*nums)
    return list(chain(*map(get_numbers, s.split(';'))))


def append_doc(namedtuple_cls, doc):
    """Append a documentation to a namedtuple

    Parameters
    ----------
    namedtuple_cls: type
        The type that has been created with :func:`collections.namedtuple`
    doc: str
        The documentation docstring"""
    if six.PY3:
        namedtuple_cls.__doc__ += '\n' + doc
        return namedtuple_cls
    else:
        class DocNamedTuple(namedtuple_cls):
            __doc__ = namedtuple_cls.__doc__ + '\n' + doc
            __slots__ = ()
        DocNamedTuple.__name__ = namedtuple_cls.__name__
        return DocNamedTuple



# -*- coding: utf-8 -*-
"""Energy consumption module of the iucm package

This module, together with the `dist_numerator` package contains the necessary
functions for computing the energy consumption of a city. The main API function
is the :func:`energy_consumption` function. Note that the :data:`OWN` value is
set to the value for Stuttgart, Germany. You should change it if you want to
model another city. The parameters of this module come from [LeNechet2012]_.

References
----------
.. [LeNechet2012] Le Néchet, Florent. "Urban spatial structure, daily
   mobility and energy consumption: a study of 34 european cities."
   Cybergeo: European Journal of Geography (2012)."""
from __future__ import division
from scipy import stats
import numpy as np
from numpy.linalg import lstsq
from iucm.dist import dist
import iucm.utils as utils
from collections import namedtuple

docstrings = utils.docstrings


if np.__version__ < '1.14':
    rcond = -1
else:
    rcond = None


#: number of cars per 100 people. The default is 0, i.e. the value is ignored.
#: Another possible value that has been previously used might be 37.7, the
#: value for Stuttgart
OWN = 0

#: Intercept of energy consumption
K = -346.5

# ---- The weights for calculating the energy consumption ----
#: weight for car owner ship on energy consumption
wOWN = 17.36

#: weight for average distance of two individuals on energy consumption
wDIST = 279.0

#: weight for rank-size rule slope on energy consumption
wRSS = -9343.0

#: weight for entropy on energy consumption
wENTROP = 21700.0

EnVariables = utils.append_doc(
    namedtuple('EnVariables', ['k', 'dist', 'entrop', 'rss', 'own']),
    """A tuple containing the values that setup the energy consumption

    Parameters
    ----------
    k: float
        Intercept of energy consumption
    dist: float
        Average distance between two individuals
    entrop: float
        Entropy
    rss: float
        Rank-Size Slope
    own: float
        weight for car owner ship on energy consumption

    See Also
    --------
    iucm.model.Output
    iucm.model.Output2D
    iucm.model.PopulationModel.state
    iucm.model.PopulationModel.allocate_output
    """)


#: The weights for the variables to calculate the energy_consumption from
#: the multiple regression after [LeNechet2012]_
weights_LeNechet = EnVariables(k=K, dist=wDIST, rss=wRSS, entrop=wENTROP,
                               own=wOWN)

#: The standard errors of the weights used in [LeNechet2012]_ (obtained through
#: private communication via R. Cremades)
std_err_LeNechet = EnVariables(k=10500., own=4.696, dist=74.88, rss=2776.,
                               entrop=9172.)


def random_weights(weights):
    """Draw random weights

    This functions draws random weights and fills the arrays in the given
    `weights` with them. Weights are drawn using normal distributions defined
    through the :attr:`weights_LeNechet` and :attr:`std_err_LeNechet`.

    Parameters
    ----------
    weights: EnVariables
        The arrays to fill

    Notes
    -----
    `weights` are modified inplace!"""
    for arr, loc, scale in zip(weights, weights_LeNechet, std_err_LeNechet):
        arr[:] = np.random.normal(loc, scale, arr.shape)


def rss(population):
    """
    Compute the Rank-Size slope coefficient

    The rank-size coefficient :math:`a > 0` is calculated through a linear fit
    after [LeNechet2012]_ with

    .. math::

        \\log\\frac{p_k}{p_1} = -a \\log{k}

    where :math:`p_k` is the :population of the math:`k`-th ranking cell.

    Parameters
    ----------
    population: 1D np.ndarray
        The population data (must not contain 0!)

    Returns
    -------
    float
        The rank-size coefficient
    """
    k = stats.rankdata(population, method='dense')[:, np.newaxis]
    k[:, :] = k.max() - k[:, :] + 1
    k[:, :] = np.log(k)
    pmax = population.max()
    y_fit = np.log(population / pmax)
    ret = lstsq(k, y_fit, rcond=rcond)[0][0]
    return np.abs(ret)


def entrop(population, size):
    """
    Compute the entropy of a city

    Compute the entropy of a city after [LeNechet2012]_ via

    .. math::

        ENTROP = \\frac{
            \\sum_{i=1}^{size}\\frac{p_i}{P_{sum}} \\log\\frac{p_i}{P_{sum}}}{
            \log size}

    Parameters
    ----------
    population: 1D np.ndarray
        The population data (must not contain 0!)
    size: int
        The original size of the `population` data (including 0)

    Returns
    -------
    float
        The entropy value
    """
    inv_pop = 1 / population.sum()
    s = inv_pop / np.log10(size)
    return -(population * np.log10(population * inv_pop)).sum() * s


def energy_consumption(population, x, y, dist0=-1, slicer=None, indices=None,
                       increase=None, weights=weights_LeNechet):
    """
    Compute the energy consumption of a city

    Compute the energy consumption according to [LeNechet2012]_ via

    .. math::

        E = -346 + 17.4 \\cdot OWN + 279\\cdot DIST
                 - 9340\\cdot RSS + 21700\\cdot ENTROP

    Parameters
    ----------
    population: 1D np.ndarray
        The 1D population data
    x: 1D np.ndarray
        The x coordinates information for each cell in `population` in
        kilometers
    y: 1D np.ndarray
        The y coordinates information for each cell in `population` in
        kilometers
    dist0: float, optional
        The previous average distance between two individuals (see
        :func:`iucm.dist.dist` function). Speeds up the computation
        significantly
    slicer: :class:`slice` or boolean array, optional
        The slicer that can be use to access the changed cells specified by
        `increase`
    indices: 1D np.ndarray of dtype int, optional
        The indices corresponding to the increase in `increase`
    increase: 1D np.ndarray, optional
        The changed population which will be added on `population`. Specifying
        this and `dist0` speeds up the computation significantly instead of
        using the `population` alone. Note that you must then also specify
        `slicer` and `indices`
    weights: EnVariables
        The multiple regression coefficients (weights) for the calculating the
        energy consumption. If not given, the (0-dimensional) weights after
        [LeNechet2012]_ (:attr:`weights_LeNechet`, see above equation) are
        used.


    Returns
    -------
    np.ndarray of dtype float
        The energy consumption. The shape of the array depends on the given
        `weights`
    float
        The average distance between two individuals (DIST)
    float
        The entropy ENTROP
    float
        The rank-size-slope RSS

    See Also
    --------
    OWN
    entrop
    rss
    dist_numerator
    """
    if increase is not None:
        increased = population.copy()
        increased[slicer] += increase
    else:
        increased = population
    increased = increased[increased > 0]
    if x.size != population.size:
        # 1d-coordinates
        x, y = [a.astype(np.float64).ravel() for a in np.meshgrid(x, y)]
    elif x.ndim == 2:
        x = x.ravel()
        y = y.ravel()
    # round the DIST variable because otherwise you may get numerical
    # instabilities after multiple steps
    DIST = np.round(dist(population, x, y, dist0=dist0, indices=indices,
                         increase=increase), 5)
    ENTROP = entrop(increased, population.size)
    try:
        RSS = rss(increased)
    except RuntimeError:
        return np.inf, DIST, ENTROP, 0
    return (
        # energy consumption
        _calculate_en([DIST, ENTROP, RSS], weights, False),
        # other returns
        DIST, ENTROP, RSS)


def _calculate_en(dist_entrop_rss,
                  weights=np.array(weights_LeNechet)[:, np.newaxis],
                  squeeze=True):
    """Calculate the energy consumption

    Parameters
    ----------
    dist_entrop_rss: 1D np.ndarray of dtype float and shape (3, )
        DIST, ENTROP and RSS
    weights: 2D np.ndarray of shape (5, N)
        The weights to use (by default, :attr:`weights_LeNechet`). They must
        correspond to the attributes in the :class:`EnVariables` class
    squeeze: bool
        If True, squeeze the result

    Returns
    -------
    np.ndarray of shape (N, )
        The energy consumption. If ``N == 1 and squeeze``, this dimension is
        squeezed
    """
    weights = np.asarray(weights)
    if weights.ndim == 1:
        return weights[0] + weights[-1] * OWN + \
            np.sum(np.asarray(dist_entrop_rss) * weights[1:-1])
    else:
        ret = weights[0] + weights[-1] * OWN + \
            np.sum(np.asarray(dist_entrop_rss)[:, np.newaxis] * weights[1:-1],
                   axis=0)
    if squeeze:
        return np.squeeze(ret)
    else:
        return ret



from setuptools import setup, find_packages, Extension
import numpy as np
import os.path as osp

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True


def readme():
    with open('README.rst') as f:
        return f.read()


ext = '.pyx' if USE_CYTHON else '.c'


extensions = [Extension('iucm._dist',
                        sources=[osp.join("iucm", "_dist" + ext)])]


if USE_CYTHON:
    extensions = cythonize(extensions)


# read the __version__ from version.py
with open(osp.join('iucm', 'version.py')) as f:
    exec(f.read())


setup(
    name='iucm',
    version=__version__,
    description='Integrated Urban Complexity Model',
    long_description=readme(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v3 or later '
        '(GPLv3+)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
    ],
    keywords='urban-planning climate energy python modeling',
    ext_modules=extensions,
    packages=find_packages(exclude=['docs', 'tests*', 'examples']),
    include_dirs=[np.get_include()],
    install_requires=[
        'xarray',
        'scipy',
        'netCDF4',
        'psyplot',
        'model_organization',
        ],
    entry_points={'console_scripts': ['iucm=iucm.main:main']},
    zip_safe=False,
    data_files=[("", ["LICENSE"])],
    url='https://github.com/Chilipp/iucm',
    author='Philipp S. Sommer and Roger Cremades',
    author_email='philipp.sommer@unil.ch',
    )



"""IUCM - The Integrated Urban Complexity Model

A model to increase the population in an efficient way with respect to
transportation energy
"""

from iucm.version import __version__

__author__ = "Philipp Sommer and Roger Cremades"
__maintainer__ = "Philipp Sommer"
__licence__ = "GPLv3"
__email__ = "philipp.sommer@unil.ch"
__copyright__ = "Copyright 2016, Philipp Sommer and Roger Cremades"



"""Module for calculating the average distance between two individuals"""
from iucm._dist import dist as _dist


def dist(population, x, y, dist0=-1, indices=None, increase=None):
    """Calculate the average distance between two individuals

    This function calculates

    .. math::

        d = \\frac{\\sum_{i,j=1}^N d_{ij} P_i P_j}{P_{tot} (P_{tot}-1)}

    after [LeNechet2012]_, where :math:`P_i` is the population of the i-th grid
    cell, :math:`P_{tot}` is the total population and :math:`d_{ij}` is the
    distance between the i-th and j-th grid cell

    Parameters
    ----------
    population: 1D np.ndarray of dtype float
        The population data for each grid cell. If ``dist0>=0`` then this
        array must hold the population of the previous step corresponding to
        `dist0`. Otherwise it should be the real population.
    x: 1D np.ndarray of dtype float
        The x-coordinates for each element in `population`
    y: 1D np.ndarray of dtype float
        The y-coordinates for each element in `population`
    dist0: float, optional
        The previous average distance between individuals. If this is given,
        `increase` and `indices` must not be None and `population` must
        represent the population of the previous step. Specifying this value,
        significantly speeds up the calculation.
    indices: 1D np.ndarray of dtype int
        The indices of the grid cells, where the population has been increased.
    increase: 1D np.ndarray of dtype float
        The increase in the grid cells corresponding to the given `indices`

    Returns
    -------
    float
        The average distance between two individuals.
    """
    return _dist(population, x, y, dist0, indices, increase)
